""" This file has been used as-is and with permission from the authors/maintainers of
NASLib, at https://github.com/automl/NASLib."""

from __future__ import print_function

import sys
import logging
import argparse
import torchvision.datasets as dset
from torch.utils.data import Dataset
from sklearn import metrics
from scipy import stats

from copy import copy
from collections import OrderedDict

import random
import os
import os.path
import shutil
from functools import wraps, partial
from pathlib import Path

import numpy as np
import torch
import torchvision.transforms as transforms
import yaml

from fvcore.common.checkpoint import Checkpointer as fvCheckpointer
from fvcore.common.config import CfgNode

cat_channels = partial(torch.cat, dim=1)

logger = logging.getLogger(__name__)


def get_project_root() -> Path:
    """
    Returns the root path of the project.
    """
    return Path(__file__).parent.parent


def iter_flatten(iterable):
    """
    Flatten a potentially deeply nested python list
    """
    # taken from https://rightfootin.blogspot.com/2006/09/more-on-python-flatten.html
    it = iter(iterable)
    for e in it:
        if isinstance(e, (list, tuple)):
            for f in iter_flatten(e):
                yield f
        else:
            yield e


def default_argument_parser():
    """
    Returns the argument parser with the default options.

    Inspired by the implementation of FAIR's detectron2
    """

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--config-file",
        default="{}/benchmarks/predictors/predictor_config.yaml".format(
            get_project_root()
        ),
        metavar="FILE",
        help="path to config file",
    )
    # parser.add_argument("--config-file", default="{}/defaults/darts_defaults.yaml".format(get_project_root()), metavar="FILE", help="path to config file")
    parser.add_argument(
        "--eval-only", action="store_true", help="perform evaluation only"
    )
    parser.add_argument("--seed", default=0, help="random seed")
    parser.add_argument(
        "--resume", action="store_true", help="Resume from last checkpoint"
    )
    parser.add_argument(
        "--model-path", type=str, default=None, help="Path to saved model weights"
    )
    parser.add_argument(
        "--world-size",
        default=1,
        type=int,
        help="number of nodes for distributed training",
    )
    parser.add_argument(
        "--rank", default=0, type=int, help="node rank for distributed training"
    )
    parser.add_argument("--gpu", default=None, type=int, help="GPU id to use.")
    parser.add_argument(
        "--dist-url",
        default="tcp://127.0.0.1:8888",
        type=str,
        help="url used to set up distributed training",
    )
    parser.add_argument(
        "--dist-backend", default="nccl", type=str, help="distributed backend"
    )
    parser.add_argument(
        "--multiprocessing-distributed",
        action="store_true",
        help="Use multi-processing distributed training to launch "
             "N processes per node, which has N GPUs. This is the "
             "fastest way to use PyTorch for either single node or "
             "multi node data parallel training",
    )
    parser.add_argument(
        "opts",
        help="Modify config options using the command-line",
        default=None,
        nargs=argparse.REMAINDER,
    )
    return parser


def parse_args(parser=default_argument_parser(), args=sys.argv[1:]):
    if "-f" in args:
        args = args[2:]
    return parser.parse_args(args)


def pairwise(iterable):
    """
    Iterate pairwise over list.

    from https://stackoverflow.com/questions/5389507/iterating-over-every-two-elements-in-a-list
    """
    "s -> (s0, s1), (s2, s3), (s4, s5), ..."
    a = iter(iterable)
    return zip(a, a)


def get_config_from_args(args=None, config_type="nas"):
    """
    Parses command line arguments and merges them with the defaults
    from the config file.

    Prepares experiment directories.

    Args:
        args: args from a different argument parser than the default one.
    """

    if config_type == "nas":
        # load the default base
        with open(
            os.path.join(get_project_root(), "defaults", "darts_defaults.yaml")
        ) as f:
            config = CfgNode.load_cfg(f)
    elif config_type == "predictor":
        # load the default base
        with open(
            os.path.join(
                get_project_root(), "benchmarks/predictors", "predictor_config.yaml"
            )
        ) as f:
            config = CfgNode.load_cfg(f)
    elif config_type == "bbo-bs":
        # load the default base
        with open(
            os.path.join(
                get_project_root(), "benchmarks/bbo", "discrete_config.yaml"
            )
        ) as f:
            config = CfgNode.load_cfg(f)
    elif config_type == "nas_predictor":
        # load the default base
        # with open(os.path.join(get_project_root(), 'benchmarks/nas_predictors', 'nas_predictor_config.yaml')) as f:
        with open(
            os.path.join(
                get_project_root(), "benchmarks/nas_predictors", "discrete_config.yaml"
            )
        ) as f:
            config = CfgNode.load_cfg(f)
    elif config_type == "oneshot":
        with open(
            os.path.join(
                get_project_root(),
                "benchmarks/nas_predictors",
                "nas_predictor_config.yaml",
            )
        ) as f:
            config = CfgNode.load_cfg(f)
    elif config_type == "statistics":
        # load the default base
        with open(
            os.path.join(
                get_project_root(), "benchmarks/statistics", "statistics_config.yaml"
            )
        ) as f:
            config = CfgNode.load_cfg(f)

    if args is None:
        args = parse_args()
    logger.info("Command line args: {}".format(args))

    # Override file args with ones from command line
    try:
        for arg, value in pairwise(args.opts):
            if "." in arg:
                arg1, arg2 = arg.split(".")
                config[arg1][arg2] = type(config[arg1][arg2])(value)
            else:
                config[arg] = value

        config.eval_only = args.eval_only
        config.resume = args.resume
        config.model_path = args.model_path
        if config_type != "nas_predictor":
            config.seed = args.seed

        # load config file
        config.set_new_allowed(True)
        config.merge_from_file(args.config_file)
        config.merge_from_list(args.opts)
    except AttributeError:
        for arg, value in pairwise(args):
            config[arg] = value

    # prepare the output directories
    if config_type == "nas":
        # config.seed = args.seed
        config.search.seed = config.seed
        # config.optimizer = args.optimizer
        config.evaluation.world_size = args.world_size
        config.gpu = config.search.gpu = config.evaluation.gpu = args.gpu
        config.evaluation.rank = args.rank
        config.evaluation.dist_url = args.dist_url
        config.evaluation.dist_backend = args.dist_backend
        config.evaluation.multiprocessing_distributed = args.multiprocessing_distributed
        config.save = "{}/{}/{}/{}/{}".format(
            config.out_dir, config.search_space, config.dataset, config.optimizer,
            config.seed
        )
    elif config_type == "bbo-bs":
        # config.seed = args.seed
        config.search.seed = config.seed
        # config.optimizer = args.optimizer
        config.evaluation.world_size = args.world_size
        config.gpu = config.search.gpu = config.evaluation.gpu = args.gpu
        config.evaluation.rank = args.rank
        config.evaluation.dist_url = args.dist_url
        config.evaluation.dist_backend = args.dist_backend
        config.evaluation.multiprocessing_distributed = args.multiprocessing_distributed
        config.save = "{}/{}/{}/{}/config_{}/{}".format(
            config.out_dir, config.search_space, config.dataset, config.optimizer,
            config.config_id, config.seed
        )


    elif config_type == "predictor":
        if config.predictor == "lcsvr" and config.experiment_type == "vary_train_size":
            config.save = "{}/{}/{}/{}_train/{}".format(
                config.out_dir,
                config.dataset,
                "predictors",
                config.predictor,
                config.seed,
            )
        elif config.predictor == "lcsvr" and config.experiment_type == "vary_fidelity":
            config.save = "{}/{}/{}/{}_fidelity/{}".format(
                config.out_dir,
                config.dataset,
                "predictors",
                config.predictor,
                config.seed,
            )
        else:
            config.save = "{}/{}/{}/{}/{}".format(
                config.out_dir,
                config.dataset,
                "predictors",
                config.predictor,
                config.seed,
            )
    elif config_type == "nas_predictor":
        config.search.seed = config.seed
        config.save = "{}/{}/{}/{}/{}/{}".format(
            config.out_dir,
            config.dataset,
            "nas_predictors",
            config.search_space,
            config.search.predictor_type,
            config.seed,
        )
    elif config_type == "oneshot":
        config.save = "{}/{}/{}/{}/{}/{}".format(
            config.out_dir,
            config.dataset,
            "nas_predictors",
            config.search_space,
            config.search.predictor_type,
            config.seed,
        )
    elif config_type == "statistics":
        config.save = "{}/{}/{}/{}/{}".format(
            config.out_dir,
            config.search_space,
            config.dataset,
            "statistics",
            config.seed,
        )
    else:
        print("invalid config type in utils/utils.py")

    config.data = "{}/data".format(get_project_root())

    create_exp_dir(config.save)
    create_exp_dir(config.save + "/search")  # required for the checkpoints
    create_exp_dir(config.save + "/eval")

    return config


def get_train_val_loaders(config, mode):
    """
    Constructs the dataloaders and transforms for training, validation and test data.
    """
    data = config.data
    dataset = config.dataset
    seed = config.search.seed
    config = config.search if mode == "train" else config.evaluation
    if dataset == "cifar10":
        train_transform, valid_transform = _data_transforms_cifar10(config)
        train_data = dset.CIFAR10(
            root=data, train=True, download=True, transform=train_transform
        )
        test_data = dset.CIFAR10(
            root=data, train=False, download=True, transform=valid_transform
        )
    elif dataset == "cifar100":
        train_transform, valid_transform = _data_transforms_cifar100(config)
        train_data = dset.CIFAR100(
            root=data, train=True, download=True, transform=train_transform
        )
        test_data = dset.CIFAR100(
            root=data, train=False, download=True, transform=valid_transform
        )
    elif dataset == "svhn":
        train_transform, valid_transform = _data_transforms_svhn(config)
        train_data = dset.SVHN(
            root=data, split="train", download=True, transform=train_transform
        )
        test_data = dset.SVHN(
            root=data, split="test", download=True, transform=valid_transform
        )
    elif dataset == "ImageNet16-120":
        from naslib.utils.DownsampledImageNet import ImageNet16

        train_transform, valid_transform = _data_transforms_ImageNet_16_120(config)
        data_folder = f"{data}/{dataset}"
        train_data = ImageNet16(
            root=data_folder,
            train=True,
            transform=train_transform,
            use_num_of_class_only=120,
        )
        test_data = ImageNet16(
            root=data_folder,
            train=False,
            transform=valid_transform,
            use_num_of_class_only=120,
        )
    else:
        raise ValueError("Unknown dataset: {}".format(dataset))

    num_train = len(train_data)
    indices = list(range(num_train))
    split = int(np.floor(config.train_portion * num_train))

    train_queue = torch.utils.data.DataLoader(
        train_data,
        batch_size=config.batch_size,
        sampler=torch.utils.data.sampler.SubsetRandomSampler(indices[:split]),
        pin_memory=True,
        num_workers=0,
        worker_init_fn=np.random.seed(seed),
    )

    valid_queue = torch.utils.data.DataLoader(
        train_data,
        batch_size=config.batch_size,
        sampler=torch.utils.data.sampler.SubsetRandomSampler(indices[split:num_train]),
        pin_memory=True,
        num_workers=0,
        worker_init_fn=np.random.seed(seed),
    )

    test_queue = torch.utils.data.DataLoader(
        test_data,
        batch_size=config.batch_size,
        shuffle=False,
        pin_memory=True,
        num_workers=0,
        worker_init_fn=np.random.seed(seed),
    )

    return train_queue, valid_queue, test_queue, train_transform, valid_transform


def _data_transforms_cifar10(args):
    CIFAR_MEAN = [0.49139968, 0.48215827, 0.44653124]
    CIFAR_STD = [0.24703233, 0.24348505, 0.26158768]

    train_transform = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
        ]
    )
    if args.cutout:
        train_transform.transforms.append(Cutout(args.cutout_length, args.cutout_prob))

    valid_transform = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
        ]
    )
    return train_transform, valid_transform


def _data_transforms_svhn(args):
    SVHN_MEAN = [0.4377, 0.4438, 0.4728]
    SVHN_STD = [0.1980, 0.2010, 0.1970]

    train_transform = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(SVHN_MEAN, SVHN_STD),
        ]
    )
    if args.cutout:
        train_transform.transforms.append(Cutout(args.cutout_length, args.cutout_prob))

    valid_transform = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize(SVHN_MEAN, SVHN_STD),
        ]
    )
    return train_transform, valid_transform


def _data_transforms_cifar100(args):
    CIFAR_MEAN = [0.5071, 0.4865, 0.4409]
    CIFAR_STD = [0.2673, 0.2564, 0.2762]

    train_transform = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
        ]
    )
    if args.cutout:
        train_transform.transforms.append(Cutout(args.cutout_length, args.cutout_prob))

    valid_transform = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize(CIFAR_MEAN, CIFAR_STD),
        ]
    )
    return train_transform, valid_transform


def _data_transforms_ImageNet_16_120(args):
    IMAGENET16_MEAN = [x / 255 for x in [122.68, 116.66, 104.01]]
    IMAGENET16_STD = [x / 255 for x in [63.22, 61.26, 65.09]]

    train_transform = transforms.Compose(
        [
            transforms.RandomCrop(16, padding=2),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(IMAGENET16_MEAN, IMAGENET16_STD),
        ]
    )
    if args.cutout:
        train_transform.transforms.append(Cutout(args.cutout_length, args.cutout_prob))

    valid_transform = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize(IMAGENET16_MEAN, IMAGENET16_STD),
        ]
    )
    return train_transform, valid_transform


class TensorDatasetWithTrans(Dataset):
    """
    TensorDataset with support of transforms.
    """

    def __init__(self, tensors, transform=None):
        assert all(tensors[0].size(0) == tensor.size(0) for tensor in tensors)
        self.tensors = tensors
        self.transform = transform

    def __getitem__(self, index):
        x = self.tensors[0][index]

        if self.transform:
            x = self.transform(x)

        y = self.tensors[1][index]

        return x, y

    def __len__(self):
        return self.tensors[0].size(0)


def set_seed(seed):
    """
    Set the seeds for all used libraries.
    """
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.enabled = True
        torch.backends.cudnn.deterministic = True
        torch.cuda.manual_seed_all(seed)


def get_last_checkpoint(config, search=True):
    """
    Finds the latest checkpoint in the experiment directory.

    Args:
        config (AttrDict): The config from config file.
        search (bool): Search or evaluation checkpoint

    Returns:
        (str): The path to the latest checkpoint file.
    """
    try:
        path = os.path.join(
            config.save, "search" if search else "eval", "last_checkpoint"
        )
        with open(path, "r") as f:
            checkpoint_name = f.readline()
        return os.path.join(
            config.save, "search" if search else "eval", checkpoint_name
        )
    except:
        return ""


def accuracy(output, target, topk=(1,)):
    """
    Calculate the accuracy given the softmax output and the target.
    """
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].reshape(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


def count_parameters_in_MB(model):
    """
    Returns the model parameters in mega byte.
    """
    return (
        np.sum(
            np.prod(v.size())
            for name, v in model.named_parameters()
            if "auxiliary" not in name
        )
        / 1e6
    )


def log_args(args):
    """
    Log the args in a nice way.
    """
    for arg, val in args.items():
        logger.info(arg + "." * (50 - len(arg) - len(str(val))) + str(val))


def create_exp_dir(path):
    """
    Create the experiment directories.
    """
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)
    logger.info("Experiment dir : {}".format(path))


def cross_validation(
    xtrain, ytrain, predictor, split_indices, score_metric="kendalltau"
):
    validation_score = []

    for train_indices, validation_indices in split_indices:
        xtrain_i = [xtrain[j] for j in train_indices]
        ytrain_i = [ytrain[j] for j in train_indices]
        xval_i = [xtrain[j] for j in train_indices]
        yval_i = [ytrain[j] for j in train_indices]

        predictor.fit(xtrain_i, ytrain_i)
        ypred_i = predictor.query(xval_i)

        # If the predictor is an ensemble, take the mean
        if len(ypred_i.shape) > 1:
            ypred_i = np.mean(ypred_i, axis=0)

        # use Pearson correlation to be the metric -> maximise Pearson correlation
        if score_metric == "pearson":
            score_i = np.abs(np.corrcoef(yval_i, ypred_i)[1, 0])
        elif score_metric == "mae":
            score_i = np.mean(abs(ypred_i - yval_i))
        elif score_metric == "rmse":
            score_i = metrics.mean_squared_error(yval_i, ypred_i, squared=False)
        elif score_metric == "spearman":
            score_i = stats.spearmanr(yval_i, ypred_i)[0]
        elif score_metric == "kendalltau":
            score_i = stats.kendalltau(yval_i, ypred_i)[0]
        elif score_metric == "kt_2dec":
            score_i = stats.kendalltau(yval_i, np.round(ypred_i, decimals=2))[0]
        elif score_metric == "kt_1dec":
            score_i = stats.kendalltau(yval_i, np.round(ypred_i, decimals=1))[0]

        validation_score.append(score_i)

    return np.mean(validation_score)


def generate_kfold(n, k):
    """
    Input:
        n: number of training examples
        k: number of folds
    Returns:
        kfold_indices: a list of len k. Each entry takes the form
        (training indices, validation indices)
    """
    assert k >= 2
    kfold_indices = []

    indices = np.array(range(n))
    fold_size = n // k

    fold_indices = [indices[i * fold_size: (i + 1) * fold_size] for i in range(k - 1)]
    fold_indices.append(indices[(k - 1) * fold_size:])

    for i in range(k):
        training_indices = [fold_indices[j] for j in range(k) if j != i]
        validation_indices = fold_indices[i]
        kfold_indices.append((np.concatenate(training_indices), validation_indices))

    return kfold_indices


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class AverageMeterGroup:
    """Average meter group for multiple average meters, ported from Naszilla repo."""

    def __init__(self):
        self.meters = OrderedDict()

    def update(self, data, n=1):
        for k, v in data.items():
            if k not in self.meters:
                self.meters[k] = NamedAverageMeter(k, ":4f")
            self.meters[k].update(v, n=n)

    def __getattr__(self, item):
        return self.meters[item]

    def __getitem__(self, item):
        return self.meters[item]

    def __str__(self):
        return "  ".join(str(v) for v in self.meters.values())

    def summary(self):
        return "  ".join(v.summary() for v in self.meters.values())


class NamedAverageMeter:
    """Computes and stores the average and current value, ported from naszilla repo"""

    def __init__(self, name, fmt=":f"):
        """
        Initialization of AverageMeter
        Parameters
        ----------
        name : str
            Name to display.
        fmt : str
            Format string to print the values.
        """
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = "{name} {val" + self.fmt + "} ({avg" + self.fmt + "})"
        return fmtstr.format(**self.__dict__)

    def summary(self):
        fmtstr = "{name}: {avg" + self.fmt + "}"
        return fmtstr.format(**self.__dict__)


class AverageMeter(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.avg = 0
        self.sum = 0
        self.cnt = 0

    def update(self, val, n=1):
        self.sum += val * n
        self.cnt += n
        self.avg = self.sum / self.cnt


class Cutout(object):
    def __init__(self, length, prob=1.0):
        self.length = length
        self.prob = prob

    def __call__(self, img):
        if np.random.binomial(1, self.prob):
            h, w = img.size(1), img.size(2)
            mask = np.ones((h, w), np.float32)
            y = np.random.randint(h)
            x = np.random.randint(w)

            y1 = np.clip(y - self.length // 2, 0, h)
            y2 = np.clip(y + self.length // 2, 0, h)
            x1 = np.clip(x - self.length // 2, 0, w)
            x2 = np.clip(x + self.length // 2, 0, w)

            mask[y1:y2, x1:x2] = 0.0
            mask = torch.from_numpy(mask)
            mask = mask.expand_as(img)
            img *= mask
        return img


from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Tuple
from fvcore.common.file_io import PathManager
import os


class Checkpointer(fvCheckpointer):
    def load(self, path: str, checkpointables: Optional[List[str]] = None) -> object:
        """
        Load from the given checkpoint. When path points to network file, this
        function has to be called on all ranks.
        Args:
            path (str): path or url to the checkpoint. If empty, will not load
                anything.
            checkpointables (list): List of checkpointable names to load. If not
                specified (None), will load all the possible checkpointables.
        Returns:
            dict:
                extra data loaded from the checkpoint that has not been
                processed. For example, those saved with
                :meth:`.save(**extra_data)`.
        """
        if not path:
            # no checkpoint provided
            self.logger.info("No checkpoint found. Initializing model from scratch")
            return {}
        self.logger.info("Loading checkpoint from {}".format(path))
        if not os.path.isfile(path):
            path = PathManager.get_local_path(path)
            assert os.path.isfile(path), "Checkpoint {} not found!".format(path)

        checkpoint = self._load_file(path)
        incompatible = self._load_model(checkpoint)
        if (
            incompatible is not None
        ):  # handle some existing subclasses that returns None
            self._log_incompatible_keys(incompatible)

        for key in self.checkpointables if checkpointables is None else checkpointables:
            if key in checkpoint:  # pyre-ignore
                self.logger.info("Loading {} from {}".format(key, path))
                obj = self.checkpointables[key]
                try:
                    obj.load_state_dict(checkpoint.pop(key))  # pyre-ignore
                except:
                    print("exception loading")

        # return any further checkpoint data
        return checkpoint



# Note: This script requires an installation of the complete NASLib repository in order
# to work.
from naslib.search_spaces.nasbench201.graph import NasBench201SearchSpace
from naslib.search_spaces.nasbench201 import conversions
from jahs_bench.tabular.lib.naslib.utils.utils import get_project_root
from pathlib import Path
import pickle
import pandas as pd

ss = NasBench201SearchSpace()

nb201_to_ops = {
        'avg_pool_3x3': 'AvgPool1x1',
        'nor_conv_1x1': 'ReLUConvBN1x1',
        'nor_conv_3x3': 'ReLUConvBN3x3',
        'skip_connect': 'Identity',
        'none': 'Zero',
    }

def recover_ops(model_str: str):
    ops = [s for s in model_str.split("|") if s not in ["", "+"]]
    split_ops = {i: [] for i in range(3)}
    for op, i in [s.split("~") for s in ops]:
        split_ops[int(i)].append(conversions.OP_NAMES.index(nb201_to_ops[op]))

    final_ops = []
    for i in range(3):
        final_ops += split_ops[i]

    return tuple(final_ops)

pth = Path(get_project_root()) / "data"
with open(pth / "nb201_all.pickle", "rb") as fp:
    rawdata = pickle.load(fp)

df = pd.DataFrame(rawdata).transpose()
df.index = pd.MultiIndex.from_tuples([recover_ops(k) for k, v in df.iterrows()])
dfs = []
for dataset in df.columns:
    dfs.append(
        pd.DataFrame.from_records(df[dataset].values, index=df.index).assign(dataset=dataset).set_index("dataset",
                                                                                                        append=True))

df = pd.concat(dfs)
df.to_pickle(Path("/home/archit/thesis") / "nasb201_full_pandas.pkl.gz")



import json
import logging
import os
import traceback as tb
from pathlib import Path
from typing import Optional, Sequence, Iterable

import pandas as pd

from jahs_bench.tabular.lib.core import constants
from jahs_bench.tabular.lib.core.utils import DirectoryTree

_log = logging.getLogger(__name__)
metric_df_index_levels = [constants.MetricDFIndexLevels.taskid.value, constants.MetricDFIndexLevels.modelid.value,
                          constants.MetricDFIndexLevels.epoch.value]


def sorted_metric_files(metric_dir: Path, descending: bool = True) -> Sequence[Path]:
    return sorted(
        metric_dir.rglob("*.pkl.gz"),
        key=lambda f: float(f.name.rstrip(".pkl.gz")),
        reverse=descending
    )


def safe_load_df(df_pkl_files: Iterable[Path], cleanup: bool = False) -> Optional[pd.DataFrame]:
    df = None
    for pkl in df_pkl_files:
        try:
            df: pd.DataFrame = pd.read_pickle(pkl)
        except Exception as e:
            _log.info(f"DataFrame {pkl} is likely corrupted, trying next file in sequence.")
            _log.debug(f"Could not load DataFrame from {pkl} because of the following error:\n{tb.format_exc()}")
            if cleanup:
                os.remove(pkl)
        else:
            _log.info(f"Loaded Pandas DataFrame from {pkl}.")
            break

    return df


task_metadata_columns = [m for m in constants.standard_task_metrics if m != "model_config"]
modelid_lvl = constants.MetricDFIndexLevels.modelid.value


def task_df_hack(task_df: pd.DataFrame) -> pd.DataFrame:
    # TODO: Fix this hack at its source
    metadata = task_df[task_metadata_columns].droplevel(1, axis=1)
    model_configs = task_df["model_config"]
    return pd.concat({"model_config": model_configs, "metadata": metadata}, axis=1)


# TODO: Implement 'save_summary'
# TODO: Fix implementation of 'anonymize' - we still want to retain the "Epoch" index level.
def collate_task_models(taskid: int, basedir: Optional[Path] = None, dtree: Optional[DirectoryTree] = None,
                        cleanup: bool = False, save_summary: bool = False, anonymize: bool = False) \
        -> Optional[pd.DataFrame]:
    """
    Iterates over all models evaluated by a particular task, collects the available model-level metrics, and
    returns a collated version of the metrics as a Pandas DataFrame. Either a directory 'basedir' that serves as the
    base directory for an appropriate directory tree or an instantiated DirectoryTree object 'dtree' should be given
    alongside the taskid.

    :param taskid: int
        The task id within all available tasks at the given directory tree to be handled.
    :param basedir: Path-like
        The base directory where the DirectoryTree object will be rooted. Either this or 'dtree' must be given.
    :param dtree: DirectoryTree
        An instantiated DirectoryTree object defining the directory tree where the required data will be sought.
    :param cleanup: bool
        When True, also removes stray files, such as error descriptions from repeated runs, that are no longer
        relevant. Default: False
    :param save_summary: bool
        When True, a summary of the current status of the models w.r.t. the expected training configuration is saved
        in the task directory. Default: False
    :param anonymize: bool
        When True, the original task id and model id of the metric data is removed from the final, returned DataFrame.
        Default: False
    :return: pd.DataFrame or None
        If no valid metric data could be loaded from the directory tree, most likely because none of the runs were
        completed enough to generate the metrics data or all generated data was corrupt, None is returned. Otherwise, a
        pandas DataFrame containing the collated metric data of all models evaluated under this task is returned.
    """

    assert basedir is not None or dtree is not None, "One of 'basedir' or 'dtree' must be given."

    if dtree is None:
        dtree = DirectoryTree(basedir=basedir, taskid=taskid, read_only=True)
    else:
        dtree.taskid = taskid

    if not dtree.task_dir.exists():
        raise RuntimeError(f"No existing data found for task id {dtree.taskid} in the directory tree rooted at "
                           f"{dtree.basedir}.")

    with open(dtree.task_config_file) as fp:
        task_config = json.load(fp)

    train_config = task_config["train_config"]
    expected_num_epochs = train_config["epochs"]

    latest_task_metrics = sorted_metric_files(dtree.task_metrics_dir, descending=True)
    task_metrics = safe_load_df(latest_task_metrics, cleanup=cleanup)

    if task_metrics is None:
        _log.debug(f"No valid task metrics DataFrame found for task id {dtree.taskid} in the directory tree "
                  f"rooted at {dtree.basedir}")
        return None

    task_metrics = task_df_hack(task_metrics)
    assert ("metadata", modelid_lvl) in task_metrics.columns, \
        f"Unable to process task metrics DataFrame that does not have the column ('metadata', '{modelid_lvl}')."

    task_metrics = task_metrics.set_index(("metadata", modelid_lvl)).rename_axis([modelid_lvl], axis=0)
    models = dtree.existing_models
    model_dfs = []
    for m in models:
        dtree.model_idx = int(m.stem)
        latest_model_metrics = sorted_metric_files(dtree.model_metrics_dir, descending=True)
        model_metrics = safe_load_df(latest_model_metrics, cleanup=cleanup)

        if model_metrics is None:
            _log.debug(f"No valid metric DataFrame found for model id {dtree.model_idx}.")
            continue

        model_metrics.index = pd.MultiIndex.from_product([model_metrics.index, [dtree.model_idx]],
                                                         names=model_metrics.index.names + [modelid_lvl])
        model_dfs.append(model_metrics)

    if not model_dfs:
        return None
    else:
        big_model_df = pd.concat(model_dfs, axis=0)
        task_df = big_model_df.join(task_metrics, on=modelid_lvl, how="left", lsuffix="_model", rsuffix="_task")
        if anonymize:
            task_df = task_df.reset_index(drop=True)
        else:
            # TODO: Make the 'taskid' name programmatically consistent in 'assign'.
            task_df = task_df.assign(taskid=dtree.taskid).set_index(constants.MetricDFIndexLevels.taskid.value,
                                                                    append=True)
            task_df = task_df.reorder_levels(metric_df_index_levels, axis=0)

        return task_df


# TODO: Implement 'save_summary'
def collate_tasks(basedir: Optional[Path] = None, dtree: Optional[DirectoryTree] = None, cleanup: bool = False,
                  save_summary: bool = False, anonymize: bool = True) -> Optional[pd.DataFrame]:
    """
    Collate the metrics of all evaluated models in all tasks under the given directory tree, as defined by either
    the specified base directory 'basedir' or an instance of DirectoryTree passed as 'dtree'. The flag 'cleanup', when
    True, also removes stray files, such as error descriptions from repeated runs, that are no longer relevant. When
    the flag 'save_summary' is True, a summary of the current status of all the tasks w.r.t. their expected training
    configurations is saved in the base directory.

    :param basedir: Path-like
        The base directory where the DirectoryTree object will be rooted. Either this or 'dtree' must be given.
    :param dtree: DirectoryTree
        An instantiated DirectoryTree object defining the directory tree where the required data will be sought.
    :param cleanup: bool
        When True, also removes stray files, such as error descriptions from repeated runs, that are no longer
        relevant. Default: False
    :param save_summary: bool
        When True, a summary of the current status of the models w.r.t. the expected training configuration is saved
        in the task directory. A job-level summary of all the tasks is also saved in the base directory. Default: False
    :param anonymize: bool
        When True, the original task id and model id of the metric data is removed from the final, returned DataFrame.
        Default: False
    :return: pd.DataFrame or None
        If no valid metric data could be loaded from the directory tree, most likely because none of the runs were
        completed enough to generate the metrics data or all generated data was corrupt, None is returned. Otherwise, a
        pandas DataFrame containing the collated metric data of all models and all tasks evaluated under this task is
        returned.
    """

    assert basedir is not None or dtree is not None, "One of 'basedir' or 'dtree' must be given."

    if dtree is None:
        dtree = DirectoryTree(basedir=basedir, read_only=True)

    tasks = dtree.existing_tasks
    if tasks is None:
        _log.info(f"Found no existing task data sub-directories in the directory tree rooted at {dtree.basedir}")
        return None

    tasks = list(map(lambda t: int(t.stem), tasks))
    ntasks = len(tasks)
    task_dfs = []

    for i, t in enumerate(tasks, start=1):
        _log.info(f"Processing task {i}/{ntasks}")
        task_df = collate_task_models(taskid=t, dtree=dtree, cleanup=cleanup, save_summary=save_summary,
                                      anonymize=anonymize)
        if task_df is None:
            _log.debug(f"No valid metric data found for task id {t} in the directory tree rooted at {dtree.basedir}")
            continue

        task_dfs.append(task_df)

    if not task_dfs:
        _log.info(f"Found no valid metric data for any of the tasks in the directory tree rooted at {dtree.basedir}")
        return None
    else:
        big_task_df = pd.concat(task_dfs, axis=0)
        big_task_df = big_task_df.reorder_levels(metric_df_index_levels, axis=0)

        return big_task_df



""" A collection of constant values that are or may be re-used throughout the rest of the
code base. """

import enum

datasets = ["cifar10", "colorectal_histology", "fashion_mnist"]
fidelity_types = {"N": int, "W": int, "Resolution": float}
fidelity_params = tuple(fidelity_types.keys())

## The following are a number of constants that are related to the search space definition

# Names of cell ops
OP_NAMES = ['Identity', 'Zero', 'ConvBN3x3', 'ConvBN1x1', 'AvgPool1x1']

# 2-tuples, (in-node, out-node), that uniquely identify cell edges
EDGE_LIST = ((1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4))


# Available choices for activation function.

@enum.unique
class Activations(enum.Enum):
    ReLU = enum.auto()
    Hardswish = enum.auto()
    Mish = enum.auto()


# Mapping from the original NASBench-201 dataset's naming convention to NASLib's naming
# convention
nb201_to_ops = {
    'avg_pool_3x3': 'AvgPool1x1',
    'nor_conv_1x1': 'ConvBN1x1',
    'nor_conv_3x3': 'ConvBN3x3',
    'skip_connect': 'Identity',
    'none': 'Zero',
}

# Mapping from NASLib's naming conversion to the original NASBench-201 dataset's naming
# convention
ops_to_nb201 = {
    'AvgPool1x1': 'avg_pool_3x3',
    'ConvBN1x1': 'nor_conv_1x1',
    'ConvBN3x3': 'nor_conv_3x3',
    'Identity': 'skip_connect',
    'Zero': 'none',
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jahs_bench',
 'jahs_bench.lib',
 'jahs_bench.lib.core',
 'jahs_bench.surrogate',
 'jahs_bench.surrogate_training',
 'jahs_bench.tabular',
 'jahs_bench.tabular.clusterlib',
 'jahs_bench.tabular.lib',
 'jahs_bench.tabular.lib.core',
 'jahs_bench.tabular.lib.naslib',
 'jahs_bench.tabular.lib.naslib.search_spaces',
 'jahs_bench.tabular.lib.naslib.search_spaces.core',
 'jahs_bench.tabular.lib.naslib.utils',
 'jahs_bench.tabular.lib.postprocessing',
 'jahs_bench.tabular.scripts',
 'jahs_bench.tabular.search_space',
 'jahs_bench.tabular.surrogate',
 'jahs_bench_examples']

package_data = \
{'': ['*']}

install_requires = \
['ConfigSpace>=0.4.0,<0.5.0',
 'joblib>=1.1.0,<1.2.0',
 'numpy>=1.21.0,<1.22.0',
 'pandas>=1.3.0,<1.4.0',
 'pyyaml>=5,<6',
 'requests>=2.28.0,<3.0.0',
 'scikit-learn>=1.0.2,<1.1.0',
 'xgboost>=1.5.0,<1.6.0',
 'yacs>=0.1.8,<0.2.0']

setup_kwargs = {
    'name': 'jahs-bench',
    'version': '0.1.0',
    'description': 'The first collection of surrogate benchmarks for Joint Architecture and Hyperparameter Search.',
    'long_description': '# JAHS-Bench-201\n\nThe first collection of surrogate benchmarks for Joint Architecture and Hyperparameter Search (JAHS), built to also support and\nfacilitate research on multi-objective, cost-aware and (multi) multi-fidelity optimization algorithms.\n\n\n![Python versions](https://img.shields.io/badge/python-3.7%20%7C%203.8%20%7C%203.9%20%7C%203.10-informational)\n[![License](https://img.shields.io/badge/license-MIT-informational)](LICENSE)\n\nPlease see our [documentation here](https://automl.github.io/jahs_bench_201/).\n\n\n## Installation\n\nUsing pip\n\n```bash\npip install git+https://github.com/automl/jahs_bench_201.git\n```\n\nOptionally, you can download the data required to use the surrogate benchmark ahead of time with\n```bash\npython -m jahs_bench.download --target surrogates\n```\n\nTo test if the installation was successful, you can, e.g, run a minimal example with\n```bash\npython -m jahs_bench_examples.minimal\n```\nThis should randomly sample a configuration, and display both the sampled configuration and the result of querying the\nsurrogate for that configuration.\n\n## Using the Benchmark\n\n### Creating Configurations\n\nConfigurations in our Joint Architecture and Hyperparameter (JAHS) space are represented as dictionaries, e.g.,:\n\n```python\nconfig = {\n    \'Optimizer\': \'SGD\',\n    \'LearningRate\': 0.1,\n    \'WeightDecay\': 5e-05,\n    \'Activation\': \'Mish\',\n    \'TrivialAugment\': False,\n    \'Op1\': 4,\n    \'Op2\': 1,\n    \'Op3\': 2,\n    \'Op4\': 0,\n    \'Op5\': 2,\n    \'Op6\': 1,\n    \'N\': 5,\n    \'W\': 16,\n    \'Resolution\': 1.0,\n}\n```\n\nFor a full description on the search space and configurations see our [documentation](https://automl.github.io/jahs_bench_201/search_space).\n\n\n### Evaluating Configurations\n\n```python\nimport jahs_bench\n\nbenchmark = jahs_bench.Benchmark(task="cifar10", download=True)\n\n# Query a random configuration\nconfig = benchmark.sample_config()\nresults = benchmark(config, nepochs=200)\n\n# Display the outputs\nprint(f"Config: {config}")  # A dict\nprint(f"Result: {results}")  # A dict\n```\n\n\n### More Evaluation Options\n\nThe API of our benchmark enables users to either query a surrogate model (the default) or the tables of performance data, or train a\nconfiguration from our search space from scratch using the same pipeline as was used by our benchmark.\nHowever, users should note that the latter functionality requires the installation of `jahs_bench_201` with the\noptional `data_creation` component and its relevant dependencies. The relevant data can be automatically downloaded by\nour API. See our [documentation](https://automl.github.io/jahs_bench_201/usage) for details.\n\n## Benchmark Data\n\nWe provide [documentation for the performance dataset](https://automl.github.io/jahs_bench_201/performance_dataset) used to train our surrogate models and [further information on our surrogate models](https://automl.github.io/jahs_bench_201/surrogate).\n\n\n## Experiments and Evaluation Protocol\n\nSee [our experiments repository](https://github.com/automl/jahs_bench_201_experiments) and our [documentation](https://automl.github.io/jahs_bench_201/evaluation_protocol).\n\n## Leaderboards\n\nWe maintain [leaderboards](https://automl.github.io/jahs_bench_201/leaderboards) for several optimization tasks and algorithmic frameworks.\n',
    'author': 'Archit Bansal',
    'author_email': 'bansala@cs.uni-freiburg.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/automl/jahs_bench',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<=3.10',
}


setup(**setup_kwargs)



import ConfigSpace as CS
from jahs_bench.lib.core.constants import Activations

joint_config_space = CS.ConfigurationSpace("jahs_bench_config_space")
# noinspection PyPep8
joint_config_space.add_hyperparameters([
    CS.OrdinalHyperparameter("N", sequence=[1, 3, 5], default_value=1,
                             meta=dict(help="Number of cell repetitions")),
    CS.OrdinalHyperparameter("W", sequence=[4, 8, 16], default_value=4,
        meta=dict(help="The width of the first channel in the cell. Each of the "
                       "subsequent cell's first channels is twice as wide as the "
                       "previous cell's, thus, for a value 4 (default) of W, the first "
                       "channel widths are [4, 8, 16].")),
    CS.CategoricalHyperparameter("Op1", choices=list(range(5)), default_value=0,
        meta=dict(help="The operation on the first edge of the cell.")),
    CS.CategoricalHyperparameter("Op2", choices=list(range(5)), default_value=0,
        meta=dict(help="The operation on the second edge of the cell.")),
    CS.CategoricalHyperparameter("Op3", choices=list(range(5)), default_value=0,
        meta=dict(help="The operation on the third edge of the cell.")),
    CS.CategoricalHyperparameter("Op4", choices=list(range(5)), default_value=0,
        meta=dict(help="The operation on the fourth edge of the cell.")),
    CS.CategoricalHyperparameter("Op5", choices=list(range(5)), default_value=0,
        meta=dict(help="The operation on the fifth edge of the cell.")),
    CS.CategoricalHyperparameter("Op6", choices=list(range(5)), default_value=0,
        meta=dict(help="The operation on the sixth edge of the cell.")),
    CS.OrdinalHyperparameter("Resolution", sequence=[0.25, 0.5, 1.], default_value=1.,
        meta=dict(help="The sample resolution of the input images w.r.t. one side of the "
                       "actual image size, assuming square images, i.e. for a dataset "
                       "with 32x32 images, specifying a value of 0.5 corresponds to "
                       "using downscaled images of size 16x16 as inputs.")),
    CS.CategoricalHyperparameter("TrivialAugment", choices=[True, False],
        default_value=False, meta=dict(
            help="Controls whether or not TrivialAugment is used for pre-processing "
                 "data. If False (default), a set of manually chosen transforms is "
                 "applied during pre-processing. If True, these are skipped in favor of "
                 "applying random transforms selected by TrivialAugment.")),
    CS.CategoricalHyperparameter("Activation",
         choices=list(Activations.__members__.keys()), default_value="ReLU",
         meta=dict(help="Which activation function is to be used for the network. "
                        "Default is ReLU."))
])

# Add Optimizer related HyperParamters
optimizers = CS.CategoricalHyperparameter("Optimizer", choices=["SGD"],
    default_value="SGD", meta=dict(help="Which optimizer to use for training this model. "
                                        "This is just a placeholder for now, to be used "
                                        "properly in future versions."))
lr = CS.UniformFloatHyperparameter("LearningRate", lower=1e-3, upper=1e0,
    default_value=1e-1, log=True, meta=dict(
        help="The learning rate for the optimizer used during model training. In the "
             "case of adaptive learning rate optimizers such as Adam, this is the "
             "initial learning rate."))
weight_decay = CS.UniformFloatHyperparameter("WeightDecay", lower=1e-5, upper=1e-2,
    default_value=5e-4, log=True, meta=dict(help="Weight decay to be used by the "
                                                 "optimizer during model training."))

joint_config_space.add_hyperparameters([optimizers, lr, weight_decay])



import functools
import itertools
import logging
from typing import Optional

import numpy as np
import sklearn.metrics._regression as skregression

_log = logging.getLogger(__name__)


def exponential_bounds(y_true, y_pred, *, y_lim: float, argmin: Optional[float] = 0.5,
                       c: Optional[float] = None, min_val: Optional[float] = 1e-6,
                       multioutput: str = "raw_values"):
    """
    Generates an exponential boundary function for incoming predictions that can be
    used to impose regression value boundaries in the loss function. Whether this
    function generates an "upper" or "lower" bound is determined by the values of `y_lim`
    and `argmin`, wherein if `y_lim > argmin` an upper bound is generated, otherwise a
    lower bound is generated. Behaviour is undefined for `y_lim` == `argmin`.

    :param y_true: array-like
        The true labels/expected regression targets.
    :param y_pred: array-like
        The predicted regression values.
    :param y_lim: float or None
        The upper/lower limit on the regressand.
    :param argmin: float or None
        The value of y_pred at which the boundary function should decay to a very small
        value. If not given, it will be implicitly determined by the value of `c`.
    :param c: float or None
        The constant multiplier for the slope of the exponential ($e^{c * x}$),
        determines how quickly the loss value grows or decays away from the boundary.
    :param min_val: float or None
        The minimum value of the boudnary loss. If None, `c` and `argmin` must be given
        and `min-val` is automatically calculated.
    :param multioutput: str
        Corresponds with the equivalent parameter of SKLearn compatible metrics.
        Currently present for compatibility purposes only and effectively only implements
        the default "raw_metrics" feature.
    :return:
    """

    y_type, y_true, y_pred, multioutput = skregression._check_reg_targets(
        y_true, y_pred, multioutput
    )
    skregression.check_consistent_length(y_true, y_pred)

    if min_val is None or min_val == 0.:
        assert c is not None, \
            f"When no minimum loss value is specified (parameter 'min_val' - was " \
            f"{min_val}), the parameter 'c' must be given (was {c})."
    elif c is None:
        c = np.log(min_val) / (argmin - y_lim)
    else:
        raise RuntimeError("Only one of `c` and `min_val` should be specified.")

    # TODO: Check for numerical overflows and implement a solution e.g. cap this value
    loss = np.exp(c * (y_pred - y_lim))
    grad = c * loss
    hess = (c ** 2) * loss

    return grad, hess


def squared_error(y_true, y_pred, multioutput: str = "raw_values"):
    """
    Implements the squared error loss. Returns the gradient and hessian for corrected
    squared error loss i.e. 0.5 * ||y_pred - y_true||^2.

    :param y_true: array-like
        The true labels/expected regression targets.
    :param y_pred: array-like
        The predicted regression values.
    :param multioutput: str
        Corresponds with the equivalent parameter of SKLearn compatible metrics.
        Currently present for compatibility purposes only and effectively only implements
        the default "raw_metrics" feature.
    """

    y_type, y_true, y_pred, multioutput = skregression._check_reg_targets(
        y_true, y_pred, multioutput
    )
    skregression.check_consistent_length(y_true, y_pred)

    # loss = 0.5 * np.power(y_true - y_pred, 2)
    grad = y_pred - y_true
    hess = np.ones_like(y_pred)

    return grad, hess


def mix_objectives(*funcs, weights=None):
    """ Mix any number of cost functions (objectives), passed as positional arguments,
    into a new objective function that assumes that the components objectives are to be
    added together. A weighted mixture can be generated by providing a corresponding
    array-like of values to 'weights'. The individual functions will all be called using
    the same inputs - (y_true, y!°_pred). If any other arguments need to be passed to the
    inidividual functions, the functions themselves should be converted to partial
    functions using `functools.partial`. """

    if weights is None:
        n = len(funcs)
        weights = np.ones(n) / n
    else:
        weights = np.array(weights).squeeze()
        assert weights.shape[0] == len(funcs)
        assert weights.ndim == 1

    def mix_fn(y_true, y_pred):
        res = map(lambda f: f(y_true, y_pred), funcs)
        res = itertools.starmap(lambda r, w: (r[0] * w, r[1] * w), zip(res, weights))
        grad, hess = functools.reduce(lambda acc, new: (acc[0] + new[0], acc[1] + new[1]),
                                      res)
        return grad, hess

    return mix_fn



import functools
import logging
from typing import Callable

import numpy as np
import sklearn.base
import yacs.config as config
from enum import Enum

from jahs_bench.tabular.surrogate import constants
from jahs_bench.tabular.surrogate import loss

_log = logging.getLogger(__name__)


def custom_loss_function(loss_params: config.CfgNode) -> Callable:
    """ Given a configuration, construct a custom loss function. """

    func_map = {
        constants.default_base_loss_configs.exp_lower_bound.name: loss.exponential_bounds,
        constants.default_base_loss_configs.exp_upper_bound.name: loss.exponential_bounds,
        constants.default_base_loss_configs.se.name: loss.squared_error
    }

    funcs = []
    weights = []
    for f, p, w in zip(loss_params.funcs, loss_params.params, loss_params.weights):
        if not f:
            continue
        funcs.append(functools.partial(func_map[f], **p))
        weights.append(w)

    loss_func = loss.mix_objectives(*funcs, weights=weights)
    return loss_func


def load_pipeline_config(*params) -> config.CfgNode:
    """ Given a list of parameters, interprets the list as (key, value) pairs and
    converts them into an appropriate pipeline configuration. """

    pipeline_config = constants.default_pipeline_config.clone()
    pipeline_config.defrost()
    pipeline_config.set_new_allowed(True)
    pipeline_config.merge_from_list(params)

    try:
        loss_type = constants.RegressionLossFuncTypes(pipeline_config.loss)
    except ValueError as e:
        raise ValueError(
            f"The loss function for the pipeline should be a right-hand string from the "
            f"following mapping:\n{list(constants.RegressionLossFuncTypes)}") from e

    if loss_type is constants.RegressionLossFuncTypes.squared_error:
        return pipeline_config

    # Assume it's a custom loss function
    try:
        for f in pipeline_config.loss_params.funcs:
            if not f:
                continue
            _ = constants.default_base_loss_configs(f)
    except ValueError as e:
        raise ValueError(
            f"Unrecognized base loss function {f}, must be either '' or one of "
            f"{list(constants.default_base_loss_configs.__members__.keys())}") from e

    loss_params = pipeline_config.loss_params
    for i, p in enumerate(loss_params.params):
        # The parameters may be specified by passing in a filename instead
        if isinstance(p, str):
            cfg_file = Path(pipeline_config.config_dir) / p
            new_params = config.CfgNode(new_allowed=True).merge_from_file(cfg_file)
            loss_params.params[i] = new_params

    target_config = pipeline_config.target_config
    for i, p in enumerate(target_config.params):
        # The parameters may be specified by passing in a filename instead
        if isinstance(p, str):
            cfg_file = Path(pipeline_config.config_dir) / p
            new_params = config.CfgNode(new_allowed=True).merge_from_file(cfg_file)
            target_config.params = new_params

    pipeline_config.set_new_allowed(False)
    pipeline_config.freeze()
    return pipeline_config


class CustomTransformFunctions:
    @staticmethod
    def sigmoid(arr: np.array, k: float = 1.):
        assert k >= 0
        return np.power(1 + np.exp(-k * arr), -1)

    @staticmethod
    def inverse_sigmoid(arr: np.array, k: float = 1.):
        assert k > 0
        return (np.log(arr) - np.log(1 - arr)) / k


# Valid transformations for the target variable
class TargetTransforms(Enum):
    MinMax = sklearn.preprocessing.MinMaxScaler


class CustomTransforms(Enum):
    sigmoid = (CustomTransformFunctions.sigmoid, CustomTransformFunctions.inverse_sigmoid)
    inverse_sigmoid = \
        (CustomTransformFunctions.inverse_sigmoid, CustomTransformFunctions.sigmoid)


def apply_target_transform(target_config: config.CfgNode, final_estimator) -> \
    sklearn.base.BaseEstimator:
    """ Given a configuration describing how the targets should be transformed and a
    corresponding estimator, returns a MetaEstimator containing a pipeline such that the
    required transformations are applied to the targets before passed to
    'final_estimator'. The 'target_config' may describe either a single transformation or
    a chain of transformations. """

    if target_config.transform.lower() == "chain":
        # Chain the transformations by "prepending" each transformation (starting from
        # the last one) to the original estimator. This is necessary because sklearn
        # Pipelines don't support target transformations.
        estimator = final_estimator
        funcs = target_config.params["funcs"][::-1]
        params = target_config.params["params"][::-1]
        for f, p in zip(funcs, params):
            transformer = _get_single_transform(f, **p)
            estimator = sklearn.compose.TransformedTargetRegressor(
                transformer=transformer, regressor=estimator)
    else:
        transformer = _get_single_transform(target_config.transform,
                                            **target_config.params)
        estimator = sklearn.compose.TransformedTargetRegressor(
            transformer=transformer, regressor=final_estimator)

    return estimator


def _get_single_transform(name: str, **params) -> sklearn.base.BaseEstimator:
    """ Given the name of a known Transform and the relevant parameters, initialize a
    corresponding Transformer object and return it. """

    if "custom" in name:
        # Construct a custom transformer
        fname = name.split(":")[1]
        func, inv = CustomTransforms[fname].value
        return sklearn.preprocessing.FunctionTransformer(
            func=func, inverse_func=inv, kw_args=params, inv_kw_args=params)
    else:
        # Use an existing Transformer
        return TargetTransforms[name].value(**params)



"""
This script is intended to be used for cleaning up a raw metrics DataFrame and prepare it for being fed into the
surrogate model training. It, for example, filters out data from more than a specific number of epochs as well as can
be used to control whether or not data from configs with NaN values is to be included. This procedure must also account
for preventing a number of biases from being introduced into the dataset, such as sub-sampling the dataset from only a
sample of all available fidelity groups, thus destroying the underlying sampling distribution.

Workflow:
1.  Convert the data to long-format in order to make it easier to work with.
2.  Assign anonymized indices to each unique model config, as previously identified by the combination of "taskid" and
     "model_idx". Maintain a separate mapping of this.
3.  Assign a unique index to each fidelity group present in the model. Maintain a separate mapping of this.
4.  Save the completed long-format data in order to avoid repeating the above mappings and ease maintenance of
     consistency across multiple data sub-sampling calls.
5.  Filter out model configs' data as per the limits on the number of epochs - throw away excess data, check if configs
     with too few epochs are to be retained.
6.  Filter out model configs' data as per the desirability of NaN values, divergence, etc.
7.  Perform config-wise sub-sampling, if needed.
8.  Save the resultant dataset in a long-format usable for surrogate training.

"""

import argparse
import logging
from pathlib import Path
from typing import List, Tuple, Optional

import pandas as pd
import sklearn.model_selection

from jahs_bench.tabular.lib.postprocessing import metric_df_ops

_log = logging.getLogger(__name__)


def parse_cli():
    parser = argparse.ArgumentParser(
        "Clean up raw metrics data and prepare it for training a surrogate model. This includes anonymizing the index "
        "(replace taskid/model_idx tuples with a unique integer value), filtering out for the desired number of epochs "
        "within each model config, filtering out incomplete models (if needed), filtering out NaN values (if needed), "
        "and sub-sampling the data."
    )
    parser.add_argument("--data_pth", type=Path,
                        help="The path to the raw metrics DataFrame file. Must be either a full path or a path "
                             "relative to the current working directory. The name (and file extension) of this file is "
                             "used to save the long-format version at 'outdir'. When '--long' is given, this file "
                             "should instead be the aforementioned long-format version saved during a previous "
                             "iteration of this script. This must be a .pkl.gz file containing a pandas DataFrame.")
    parser.add_argument("--long", action="store_true",
                        help="When this flag is given, it is assumed that the file specified by 'data_pth' was "
                             "generated during a previous iteration of this script and is already in long-format. This "
                             "saves computation time as well as avoids consistency pitfalls when generating multiple "
                             "versions of the same dataset.")
    parser.add_argument("--outdir", type=Path, default=Path().cwd(),
                        help="Path to the directory where the cleaned data is to be stored as a pandas DataFrame. The "
                             "parent directory must already exist. Default: current working directory.")
    parser.add_argument("--output_filename", type=str, default="cleaned_metrics",
                        help="The name of the file in which the cleaned data will be stored. Appropriate file "
                             "extensions will automatically be added. Default: cleaned_metrics.")
    parser.add_argument("--epochs", type=int, default=200,
                        help="The number of epochs that a full run is expected to have lasted for. All runs that have "
                             "data for fewer than this many epochs will be considered incomplete "
                             "(consult --keep_incomplete_runs). For configs with data from more epochs than this, "
                             "the extra epochs' data will be removed.")
    parser.add_argument("--keep_incomplete_runs", action="store_true",
                        help="When this flag is given, data from incomplete runs will be not filtered out completely, "
                             "otherwise, all data for runs without the threshold number of epochs will be removed.")
    parser.add_argument("--keep_nan_configs", action="store_true",
                        help="When this flag is not given, if there are any NaN values in any of the metrics for any "
                             "given epoch of a config, all data from that config will be removed. Otherwise, only the "
                             "data for the epochs with NaN values will be removed. This is useful in certain cases, "
                             "e.g. when divergent configs' accuracies have been labelled with NaNs and all epochs "
                             "preceding the divergent epoch should be included.")
    parser.add_argument("--subsample", type=float, default=1.0,
                        help="Fraction of all available configs to be included in the cleaned data. Sub-sampling "
                             "occurs on a per-model-config basis, i.e. all epochs' data for each model config is "
                             "considered one unit for the sub-sampling. The script attempts to maintain the ratio of "
                             "model configs present in each fidelity group.")

    args = parser.parse_args()
    return args


def to_long_format(df: pd.DataFrame) -> pd.DataFrame:
    """ Converts an input DataFrame in the original metrics storage format to a long-format one. The returned DataFrame
    has three top-level columns: "sampling_index", "features", and "labels". SamplingIndex has the sub-columns "taskid"
    and "model_idx" and the corresponding values from the original index. The values of the "Epoch" level of the
    original index are moved underneath "features" instead in the sub-column "epoch" alongside all the sub-columns of
    "model_config" from the original DataFrame. All the metrics in the original DataFrame (with a name <metric>)
    recorded under a <set>: one of "train", "valid", and "test", are moved to the sub-column "<set>-<metric>". The
    sub-columns "FLOPS" and "latency" of "diagnostic" are also moved underneath "labels". Finally, "size_MB" from
    "metadata" is moved underneath "labels". The index itself is set to a simple RangeIndex. """

    # Separate out the features
    features = df.loc[:, "model_config"]
    features.loc[:, "epoch"] = df.index.get_level_values("Epoch")

    # Separate out the labels
    labels = df.loc[:, ["train", "valid", "test"]]
    labels.columns = metric_df_ops.collapse_index_names(labels.columns, nlevels=2)
    diagnostics = df.loc[:, ("diagnostic", ["FLOPS", "latency", "runtime"])]
    labels.loc[:, diagnostics.columns.get_level_values(1)] = diagnostics.values
    labels.loc[:, "size_MB"] = df.loc[:, ("metadata", "size_MB")]

    # Extract the original sampling index
    sampling_index = df.index.to_frame(index=False)
    features.index = sampling_index.index
    labels.index = sampling_index.index
    sampling_index.drop("Epoch", axis=1, inplace=True)

    combined = pd.concat({
        "sampling_index": sampling_index,
        "features": features,
        "labels": labels
    }, axis=1)

    return combined


def identify_model_configs(df: pd.DataFrame, inplace=False) -> Tuple[pd.Series, Optional[pd.Series]]:
    """
    Given a DataFrame in long-format, as generated by ´to_long_format()´, return two Series. The first Series maps each
    unique value from the column sampling_index of the input DataFrame to a unique integer ID, called model ID. The
    second maps each index value of the input DataFrame to the corresponding model ID. This second series can thus
    be used to group the rows according to individual model configs. If ´inplace=True`, only the former is returned
    whereas the input DataFrame is modified in-place by adding a new column named "model_ID" under "sampling_index" and
    None is returned for the latter.

    :param df: pandas DataFrame
    :param inplace: bool
    :return:
    """

    sampling_index: pd.DataFrame = df.sampling_index
    sampling_index_to_model_id = sampling_index.drop_duplicates().reset_index(drop=True)
    sampling_index_to_model_id.index.set_names(["model_ID"], inplace=True)
    sampling_index_to_model_id.reset_index(drop=False, inplace=True)

    new_index = sampling_index.merge(sampling_index_to_model_id)

    if inplace:
        df.loc[:, ("sampling_index", "model_ID")] = new_index["model_ID"]
        return sampling_index_to_model_id, None
    else:
        return sampling_index_to_model_id, new_index


def identify_fidelity_groups(df: pd.DataFrame, fids: List[str], inplace=False) -> Tuple[pd.Series, Optional[pd.Series]]:
    """
    Given a DataFrame in long-format, as generated by ´to_long_format()´, return two Series. The first Series maps each
    unique combination of values from the sub-columns under "features" identified by ´fids´ of the input DataFrame to a
    unique integer ID, called fidelity ID. The second maps each index value of the input DataFrame to the corresponding
    fidelity ID. This second series can thus be used to group the rows according to individual fidelity groups. If
    ´inplace=True`, only the former is returned whereas the input DataFrame is modified in-place by adding a new column
    named "fidelity_ID" under "sampling_index" and None is returned for the latter.

    :param df: pandas DataFrame
    :param fids: list of str
        A list of strings corresponding to the names of the fidelity parameters under "model_config" of the original
        metrics DataFrame and under "features" of the corresponding long-format DataFrame.
    :param inplace: bool
    :return:
    """

    fidelity_groups: pd.DataFrame = df.features[fids]
    fidelity_group_to_fidelity_id = fidelity_groups.drop_duplicates().reset_index(drop=True)
    fidelity_group_to_fidelity_id.index.set_names(["fidelity_ID"], inplace=True)
    fidelity_group_to_fidelity_id.reset_index(drop=False, inplace=True)

    new_index = fidelity_groups.merge(fidelity_group_to_fidelity_id)

    if inplace:
        df.loc[:, ("sampling_index", "fidelity_ID")] = new_index["fidelity_ID"]
        return fidelity_group_to_fidelity_id, None
    else:
        return fidelity_group_to_fidelity_id, new_index


def filter_nepochs(df: pd.DataFrame, nepochs: int, keep_incomplete: bool = False):
    """ Given a long-format metrics DataFrame, filter out only up to the given number of epochs' worth of training data
    for each model config. If ´keep_incomplete=True´, model configs with fewer than ´nepochs´ worth of data are not
    thrown away. Note: It is assumed that the epochs in the long-form table are 1-indexed, i.e. the first epoch's data
    has been marked by the value "1" for the field ´features.epoch´. """

    assert "sampling_index" in df.columns, "The input dataframe must be in a valid long-format."
    assert "model_ID" in df.sampling_index.columns, "The input dataframe must be in a valid long-format."

    _log.info("Throwing away excess epochs' data.")
    model_valid = df.features.epoch <= nepochs

    if keep_incomplete:
        return df.loc[model_valid]

    _log.info("Filtering out incomplete epochs' data.")
    existing_epoch_counts = df.groupby(("sampling_index", "model_ID"))[[("features", "epoch")]].max()
    existing_epoch_counts.columns = ["valid"]
    existing_epoch_counts.index.rename("model_ID", inplace=True)  # Get rid of the extra index level name

    models_complete = (existing_epoch_counts >= nepochs)
    models_complete = df.sampling_index.join(models_complete, on="model_ID")["valid"]
    df = df.loc[pd.concat([model_valid, models_complete], axis=1).all(axis=1)]

    return df


def subsample_df(df: pd.DataFrame, subsample_frac: float = 1.0) -> pd.DataFrame:
    """ Given a metrics DataFrame in long-format, with uniquely identifiable model IDs and fidelity IDs, selects a
    specified fraction of all available model configs' data (regardless of number of epochs in each) while maintaining
    the sampling distribution across fidelity groups. """

    if subsample_frac > 1.:
        raise ValueError(f"Sub-sampling only supported for a value in the range (0., 1.], was given {subsample_frac}.")
    elif subsample_frac == 1.:
        return df

    _log.info(f"Sub-sampling factor: {subsample_frac}, original dataset size: {df.shape[0]}")

    # Figure out which models, irrespective of how many epochs they have, should be selected
    index = df.loc[:, ("sampling_index", slice(None))].droplevel(0, axis=1)
    model_ids = index.loc[:, "model_ID"]
    fidelity_groups = index.loc[:, "fidelity_ID"]

    nunique_models = model_ids.drop_duplicates().shape[0]
    nunique_fidelity_groups = fidelity_groups.drop_duplicates().shape[0]
    counts = index.groupby("fidelity_ID").count()
    _log.info(f"Original dataset has a total of {nunique_models} unique model configs spread over "
              f"{nunique_fidelity_groups} unique fidelity groups, averaging to {counts.mean().mean():.2f} data points "
              f"in each group.")

    n_splits = int(1 / subsample_frac)
    # For shuffle=False, StratifiedGroupKFold is deterministic w.r.t. an invariant dataset (i.e. the features, labels
    # and groups must remain the same)
    splitter = sklearn.model_selection.StratifiedGroupKFold(n_splits=n_splits, shuffle=False)
    _, subsample_split = next(splitter.split(index, fidelity_groups, groups=model_ids))

    df = df.iloc[subsample_split]
    index = df.loc[:, ("sampling_index", slice(None))].droplevel(0, axis=1)
    model_ids = index.loc[:, "model_ID"]
    fidelity_groups = index.loc[:, "fidelity_ID"]

    nunique_models = model_ids.drop_duplicates().shape[0]
    nunique_fidelity_groups = fidelity_groups.drop_duplicates().shape[0]
    counts = index.groupby("fidelity_ID").count()
    _log.info(f"Sub-sampled dataset has a total of {df.shape[0]} data points, a total of {nunique_models} unique model "
              f"configs spread over {nunique_fidelity_groups} unique fidelity groups, averaging to "
              f"{counts.mean().mean():.2f} data points in each group.")
    return df


def clean(data_pth: Path, outdir: Path, output_filename: str, epochs: int, long: bool = False,
          keep_incomplete_runs: bool = False, keep_nan_configs: bool = False, subsample: float = 1.0):
    assert outdir.exists() and outdir.is_dir()
    outfile = outdir / f"{output_filename}.pkl.gz"

    _log.info(f"Loading {'long-format' if long else 'raw'} metrics data from {data_pth}")
    data: pd.DataFrame = pd.read_pickle(data_pth)
    raw_shape = data.shape
    _log.info(f"Read metrics DataFrame of shape {raw_shape}.")

    if not long:
        _log.info("Converting raw data to long-format table.")
        data = to_long_format(data)
        sampling_index_to_model_id, _ = identify_model_configs(data, inplace=True)
        fidelity_group_to_fidelity_id, _ = identify_fidelity_groups(data, fids=["N", "W", "Resolution"], inplace=True)
        data.sort_index(axis=1, inplace=True)
        long_shape = data.shape
        assert long_shape[0] == raw_shape[0], f"Unable to verify data integrity after conversion from raw table of " \
                                              f"shape {raw_shape} to long table of shape {long_shape}."
        _log.info("Successfully converted raw data to long-format.")

        _log.info(f"Saving long-format table to disk at {outdir}/long_{data_pth.name}.")
        data.to_pickle(outdir / f"long_{data_pth.name}", protocol=4)
        sampling_index_to_model_id.to_pickle(outdir / "sampling_index_to_model_id.pkl.gz", protocol=4)
        fidelity_group_to_fidelity_id.to_pickle(outdir / "fidelity_group_to_fidelity_id.pkl.gz", protocol=4)

    data = filter_nepochs(data, nepochs=epochs, keep_incomplete=keep_incomplete_runs)

    if not keep_nan_configs:
        # TODO: Handle datasets that don't include validation data
        _log.info("Handling NaN values.")
        nan_check = data.labels.isna().any(axis=1)
        nan_ind: pd.MultiIndex = data.loc[nan_check].index

        # if keep_nan_configs:
        #     # Generalize the indices to be dropped to the entire config instead of just particular epochs.
        #     nan_ind = nan_ind.droplevel("Epoch").drop_duplicates()

        data = data.drop(nan_ind)

    data = subsample_df(df=data, subsample_frac=subsample)

    _log.info(f"Saving cleaned dataset, shape: {data.shape}.")
    data.to_pickle(outfile, protocol=4)

    _log.info("Done.")


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
                        datefmt="%m/%d %H:%M:%S")
    cli_args = parse_cli()
    clean(**vars(cli_args))



import logging
from enum import Enum, unique, auto
from pathlib import Path
from typing import Optional, Union, Sequence, Tuple

import numpy as np
import pandas as pd

from jahs_bench.lib.core.configspace import joint_config_space
from jahs_bench.surrogate.model import XGBSurrogate
import jahs_bench.download

## Requires installation of the optional "data_creation" components and dependencies
try:
    from jahs_bench.tabular.sampling import run_task
    from jahs_bench.tabular.lib.core.constants import Datasets as _Tasks
    from jahs_bench.tabular.lib.core.utils import DirectoryTree, MetricLogger, \
        AttrDict
    import shutil
except ImportError as e:
    data_creation_available = False
else:
    data_creation_available = True

_log = logging.getLogger(__name__)
_log.setLevel(logging.WARNING)

@unique
class BenchmarkTypes(Enum):
    Surrogate = "surrogate"
    Table = "table"
    Live = "live"


@unique
class BenchmarkTasks(Enum):
    CIFAR10 = "cifar10"
    ColorectalHistology = "colorectal_histology"
    FashionMNIST = "fashion_mnist"


class Benchmark:
    _call_fn = None
    _surrogates = None
    _table = None

    def __init__(self, task: Union[str, BenchmarkTasks],
                 kind: Union[str, BenchmarkTypes] = BenchmarkTypes.Surrogate,
                 download: bool = True, save_dir: Union[str, Path] = "jahs_bench_data"):
        """ Load up the benchmark for querying. """

        if isinstance(task, str):
            try:
                task = BenchmarkTasks(task.lower())
            except ValueError as e:
                valid = [x.value for x in BenchmarkTasks]
                raise ValueError(f"Invalid/Unknown value of parameter 'task': '{task}'. "
                                 f"Must be one of {valid}.") from e

        if isinstance(kind, str):
            try:
                kind = BenchmarkTypes(kind)
            except ValueError as e:
                valid = [x.value for x in BenchmarkTypes]
                raise ValueError(f"Invalid/Unknown value of parameter 'kind': '{kind}'. "
                                 f"Must be one of {valid}.") from e

        self.kind = kind
        self.task = task
        self.save_dir = Path(save_dir)
        self.surrogate_dir = self.save_dir / "assembled_surrogates"
        self.table_dir = self.save_dir / "metric_data"
        self.task_dir = self.save_dir / "tasks"

        if download and kind is BenchmarkTypes.Surrogate:
            if not self.surrogate_dir.exists():
                jahs_bench.download.download_surrogates(self.save_dir)

        if download and kind is BenchmarkTypes.Table:
            if not self.table_dir.exists():
                jahs_bench.download.download_metrics(self.save_dir)

        if download and kind is BenchmarkTypes.Live:
            # TODO: Implement
            # if not self.table_dir.exists():
            #     jahs_bench.download.download_tasks(self.save_dir)
            pass

        loaders = {
            BenchmarkTypes.Surrogate: self._load_surrogate,
            BenchmarkTypes.Table: self._load_table,
            BenchmarkTypes.Live: self._load_live,
        }

        # Setup the benchmark
        loaders[kind]()

    def _load_surrogate(self):
        assert self.surrogate_dir.exists() and self.surrogate_dir.is_dir()

        model_path = self.surrogate_dir / self.task.value
        outputs = [p.name for p in model_path.iterdir() if p.is_dir()]

        self._surrogates = {}
        for o in outputs:
            self._surrogates[o] = XGBSurrogate.load(model_path / str(o))
        self._call_fn = self._benchmark_surrogate

    def _load_table(self):
        assert self.save_dir.exists() and self.save_dir.is_dir()

        table_path = self.save_dir / self.task.value
        table_names = ["train_set.pkl.gz", "valid_set.pkl.gz", "test_set.pkl.gz"]
        tables = [pd.read_pickle(table_path / n) for n in table_names]
        table = pd.concat(tables, axis=0)
        del tables

        # level_0_cols = ["features", "labels"]
        features: list = joint_config_space.get_hyperparameter_names() + ["epoch"]

        if table["features"].columns.intersection(features).size != len(features):
            raise ValueError(f"The given performance datasets at {table_path} could not "
                             f"be resolved against the known search space consisting of "
                             f"the parameters {features}")

        self._features = table["features"].columns
        self._labels: pd.Index = table["labels"].columns
        self._table_features = table.loc[:, "features"]
        self._table_labels = table.loc[:, "labels"]
        self._table_features.rename_axis("Sample ID", axis=0, inplace=True)
        self._table_features = self._table_features.reset_index()
        self._call_fn = self._benchmark_tabular

    def _load_live(self):
        self._call_fn = self._benchmark_live

    def __call__(self, config: dict, nepochs: Optional[int] = 200,
                 full_trajectory: bool = False, **kwargs):
        return self._call_fn(config=config, nepochs=nepochs,
                             full_trajectory=full_trajectory, **kwargs)

    def _benchmark_surrogate(self, config: dict, nepochs: Optional[int] = 200,
                             full_trajectory: bool = False, **kwargs) -> dict:
        assert nepochs > 0

        if full_trajectory:
            features = pd.DataFrame([config] * nepochs)
            epochs = list(range(1, nepochs+1))
            features = features.assign(epoch=epochs)
        else:
            features = pd.Series(config).to_frame().transpose()
            epochs = [nepochs]
            features.loc[:, "epoch"] = nepochs

        outputs = []
        for model in self._surrogates.values():
            outputs.append(model.predict(features))

        outputs: pd.DataFrame = pd.concat(outputs, axis=1)
        # outputs = outputs.reindex(index=epochs)
        outputs.index = epochs
        return outputs.to_dict(orient="index")

    def _benchmark_tabular(self, config: dict, nepochs: Optional[int] = 200,
                           full_trajectory: bool = False, **kwargs) -> dict:
        assert nepochs > 1
        assert self._table_features is not None and self._table_labels is not None,\
            "No performance dataset has been loaded into memory - a tabular query " \
            "cannot be made."

        query = config.copy()
        if "epoch" in query:
            query.pop("epoch")

        query_df = pd.DataFrame(query, index=list(range(1, nepochs+1)) \
                                if full_trajectory else [nepochs])
        query_df.rename_axis("epoch", axis=0, inplace=True)
        query_df = query_df.reset_index()

        check = self._features.difference(query_df.columns)
        if check.size != 0:
            raise ValueError(f"The given query has missing parameters: {check.tolist()}")

        idx = pd.merge(self._table_features, query_df, how="inner")["Sample ID"]

        if idx.size == 0:
            raise KeyError(f"Could not find any entries for the config {config} at "
                           f"{nepochs} epochs.") from e
        elif full_trajectory:
            # Return the full trajectory, but only for the first instance of this config
            # that was found.
            result = self._table_labels.loc[idx, :].iloc[:nepochs]
        else:
            # Return only the first result that was found
            result = self._table_labels.loc[idx, :].iloc[:1]

        return result.to_dict(orient="index")

    def _benchmark_live(self, config: dict, nepochs: Optional[int] = 200,
                        full_trajectory: bool = False, *,
                        train_config: Optional[dict] = None,
                        worker_dir: Optional[Path] = None, clean_tmp_files : bool = True,
                        **kwargs) -> dict:
        """
        Simple wrapper around the base benchmark data generation capabilities offered by
        tabular_sampling.distributed_nas_sampling.run_task(). Providing 'train_config'
        and 'kwargs' dicts can be used to access the full range of customizations offered
        by 'run_task()' - consults its documentation if needed. `worker_dir` can be used
        to specify a custom directory where temporary files generated during model
        training are stored. If it is not provided, a directory will be created at
        '/tmp/jahs_bench'. Disable the flag `clean_tmp_files` to retain these
        temporary files after the function call.
        """

        if not data_creation_available:
            raise RuntimeError(
                f"Cannot train the given configuration since the required modules have "
                f"not been installed. Optional component 'data_creation' of "
                f"jahs_bench must be installed in order to perform a live training "
                f"of the given configuration. Alternatively, try to query on either the "
                f"surrogate model or the performance dataset table.")

        if train_config is None:
            train_config = dict(epochs=nepochs, batch_size=256, use_grad_clipping=False,
                                split=True, warmup_epochs=0, disable_checkpointing=True,
                                checkpoint_interval_seconds=3600,
                                checkpoint_interval_epochs=50)

        basedir = (Path("/tmp") if worker_dir is None else worker_dir) / "jahs_bench"
        basedir.mkdir(exist_ok=True)
        task = _Tasks[self.task.value]

        args = {**dict(basedir=basedir, taskid=0, train_config=AttrDict(train_config),
                       dataset=task, datadir=self.task_dir, local_seed=None, global_seed=None,
                       debug=False, generate_sampling_profile=False, nsamples=1,
                       portfolio_pth=None, cycle_portfolio=False, opts=config), **kwargs}
        run_task(**args)

        dtree = DirectoryTree(basedir=basedir, taskid=0, model_idx=1, read_only=True)
        metric_pth = MetricLogger._get_sorted_metric_paths(pth=dtree.model_metrics_dir)
        metric_pth = metric_pth[-1]
        df: pd.DataFrame = pd.read_pickle(metric_pth)

        if full_trajectory:
            result = df.to_dict(orient="index")
        else:
            # It is possible that the model training diverged, so we cannot directly use
            # the value of 'nepochs' provided by the user.
            nepochs = df.index.max()
            result = df.loc[nepochs].to_dict(orient="index")

        if clean_tmp_files:
            shutil.rmtree(dtree.basedir, ignore_errors=True)

        return result

    def sample_config(self,
                      random_state: Optional[Union[int, np.random.RandomState]] = None,
                      ) -> dict:
        """ For a tabular benchmark, return a random configuration from the set of
        configurations recorded in the currently loaded dataset. Otherwise, randomly
        sample a configuration from the full search space and return it. """

        if not isinstance(random_state, np.random.RandomState):
            random_state = np.random.RandomState(random_state)

        if self.kind is BenchmarkTypes.Table:
            assert self._table_features is not None, \
                "Cannot extract random sample - the tabular benchmark could not be " \
                "properly initialized."
            index = random_state.choice(self._table_features.index)
            row = self._table_features.loc[index].drop("Sample ID")
            config = row.to_dict()
        else:
            joint_config_space.random = random_state
            config = joint_config_space.sample_configuration().get_dictionary()
            nepochs = random_state.randint(1, 200)
            config['epoch'] = nepochs

        return config


if __name__ == "__main__":
    b = Benchmark(task="cifar10", kind="surrogate")
    config = b.sample_config()
    print(f"Config: {config}")

    result = b(config, nepochs=config["epoch"])
    print(f"Result: {result}")
